--
-- newt wrapper for Euphoria
-- by Greg Haberek <ghaberek@gmail.com>
--
-- The newt windowing system is a terminal-based window and widget library designed
-- for writing applications with a simple, but user-friendly, interface. While
-- newt is not intended to provide the rich feature set advanced applications may
-- require, it has proven to be flexible enough for a wide range of applications
-- (most notably, Red Hat's installation process).
--
--
-- Documentation: Writing Programs Using newt
-- http://gnewt.sourceforge.net/tutorial.html
--

namespace newt

include "std/dll.e"
include "std/machine.e"

public constant NULL = 0

public atom libnewt = open_dll( "libnewt.so.0.52" )

public constant
    NEWT_COLORSET_ROOT              = 2,
    NEWT_COLORSET_BORDER            = 3,
    NEWT_COLORSET_WINDOW            = 4,
    NEWT_COLORSET_SHADOW            = 5,
    NEWT_COLORSET_TITLE             = 6,
    NEWT_COLORSET_BUTTON            = 7,
    NEWT_COLORSET_ACTBUTTON         = 8,
    NEWT_COLORSET_CHECKBOX          = 9,
    NEWT_COLORSET_ACTCHECKBOX       = 10,
    NEWT_COLORSET_ENTRY             = 11,
    NEWT_COLORSET_LABEL             = 12,
    NEWT_COLORSET_LISTBOX           = 13,
    NEWT_COLORSET_ACTLISTBOX        = 14,
    NEWT_COLORSET_TEXTBOX           = 15,
    NEWT_COLORSET_ACTTEXTBOX        = 16,
    NEWT_COLORSET_HELPLINE          = 17,
    NEWT_COLORSET_ROOTTEXT          = 18,
    NEWT_COLORSET_EMPTYSCALE        = 19,
    NEWT_COLORSET_FULLSCALE         = 20,
    NEWT_COLORSET_DISENTRY          = 21,
    NEWT_COLORSET_COMPACTBUTTON     = 22,
    NEWT_COLORSET_ACTSELLISTBOX     = 23,
    NEWT_COLORSET_SELLISTBOX        = 24,
    NEWT_ARG_LAST                   = (-100000),
    NEWT_ARG_APPEND                 = (-1)

public function NEWT_COLORSET_CUSTOM( integer x)
    return (30 + (x))
end function

public enum type Colors
    Colors_rootFg = 0, Colors_rootBg = 4,
    Colors_borderFg = 8, Colors_borderBg = 12,
    Colors_windowFg = 16, Colors_windowBg = 20,
    Colors_shadowFg = 24, Colors_shadowBg = 28,
    Colors_titleFg = 32, Colors_titleBg = 36,
    Colors_buttonFg = 40, Colors_buttonBg = 44,
    Colors_actButtonFg = 48, Colors_actButtonBg = 52,
    Colors_checkboxFg = 56, Colors_checkboxBg = 60,
    Colors_actCheckboxFg = 64, Colors_actCheckboxBg = 68,
    Colors_entryFg = 72, Colors_entryBg = 76,
    Colors_labelFg = 80, Colors_labelBg = 84,
    Colors_listboxFg = 88, Colors_listboxBg = 92,
    Colors_actListboxFg = 96, Colors_actListbox = 100,
    Colors_textboxFg = 104, Colors_textboxBg = 108,
    Colors_actTextboxFg = 112, Colors_actTextboxBg = 116,
    Colors_helpLineFg = 120, Colors_helpLineBg = 124,
    Colors_rootTextFg = 128, Colors_rootTextBg = 132,
    Colors_emptyScale = 136, Colors_fullScale = 140,
    Colors_disabledEntryFg = 144, Colors_disabledEntryBg = 148,
    Colors_compactButtonFg = 152, Colors_compactButtonBg = 156,
    Colors_actSelListboxFg = 160, Colors_actSelListboxBg = 164,
    Colors_selListboxFg = 168, Colors_selListboxBg = 172
end type

public enum type newtFlagsSense NEWT_FLAGS_SET, NEWT_FLAGS_RESET, NEWT_FLAGS_TOGGLE end type

public constant
    NEWT_FLAG_RETURNEXIT    = #0001, -- (1 << 0)
    NEWT_FLAG_HIDDEN        = #0002, -- (1 << 1)
    NEWT_FLAG_SCROLL        = #0004, -- (1 << 2)
    NEWT_FLAG_DISABLED      = #0008, -- (1 << 3)
--  NEWT_FLAG_NOSCROLL      = #0010, -- (1 << 4)  /* OBSOLETE      for listboxes */
    NEWT_FLAG_BORDER        = #0020, -- (1 << 5)
    NEWT_FLAG_WRAP          = #0040, -- (1 << 6)
    NEWT_FLAG_NOF12         = #0080, -- (1 << 7)
    NEWT_FLAG_MULTIPLE      = #0100, -- (1 << 8)
    NEWT_FLAG_SELECTED      = #0200, -- (1 << 9)
    NEWT_FLAG_CHECKBOX      = #0400, -- (1 << 10)
    NEWT_FLAG_PASSWORD      = #0800, -- (1 << 11) /* draw '*'  of chars in entrybox */
    NEWT_FLAG_SHOWCURSOR    = #1000, -- (1 << 12) /* Only applies to listbox for now */
    NEWT_FD_READ            = #0001, -- (1 << 0)
    NEWT_FD_WRITE           = #0002, -- (1 << 1)
    NEWT_FD_EXCEPT          = #0004, -- (1 << 2)

    NEWT_CHECKBOXTREE_UNSELECTABLE  = #1000, -- (1 << 12)
    NEWT_CHECKBOXTREE_HIDE_BOX      = #2000, -- (1 << 13)
    NEWT_CHECKBOXTREE_COLLAPSED     = 0, -- '\0'
    NEWT_CHECKBOXTREE_EXPANDED      = 1, -- '\1'
    NEWT_CHECKBOXTREE_UNSELECTED    = ' ',
    NEWT_CHECKBOXTREE_SELECTED      = '*',
    /* Backwards compatibility */
    NEWT_LISTBOX_RETURNEXIT = NEWT_FLAG_RETURNEXIT,
    NEWT_ENTRY_SCROLL       = NEWT_FLAG_SCROLL,
    NEWT_ENTRY_HIDDEN       = NEWT_FLAG_HIDDEN,
    NEWT_ENTRY_RETURNEXIT   = NEWT_FLAG_RETURNEXIT,
    NEWT_ENTRY_DISABLED     = NEWT_FLAG_DISABLED,
    NEWT_TEXTBOX_WRAP       = NEWT_FLAG_WRAP,
    NEWT_TEXTBOX_SCROLL     = NEWT_FLAG_SCROLL,
    NEWT_FORM_NOF12         = NEWT_FLAG_NOF12

public type newtComponent( object x ) return atom(x) end type

public constant
    newtDefaultColorPalette = define_c_var( libnewt, "newtDefaultColorPalette" )

public type newtCallback( object x ) return atom(x) end type
public type newtSuspendCallback( object x ) return atom(x) end type

public constant
    x_newtInit                      = define_c_func( libnewt, "newtInit", {}, C_INT ),
    x_newtFinished                  = define_c_func( libnewt, "newtFinished", {}, C_INT ),
    x_newtCls                       = define_c_proc( libnewt, "newtCls", {} ),
    x_newtResizeScreen              = define_c_proc( libnewt, "newtResizeScreen", {C_INT} ),
    x_newtWaitForKey                = define_c_proc( libnewt, "newtWaitForKey", {} ),
    x_newtClearKeyBuffer            = define_c_proc( libnewt, "newtClearKeyBuffer", {} ),
    x_newtDelay                     = define_c_proc( libnewt, "newtDelay", {C_UINT} ),
    x_newtOpenWindow                = define_c_func( libnewt, "newtOpenWindow", {C_INT,C_INT,C_UINT,C_UINT,C_POINTER}, C_INT ),
    x_newtCenteredWindow            = define_c_func( libnewt, "newtCenteredWindow", {C_UINT,C_UINT,C_POINTER}, C_INT ),
    x_newtPopWindow                 = define_c_proc( libnewt, "newtPopWindow", {} ),
    x_newtPopWindowNoRefresh        = define_c_proc( libnewt, "newtPopWindowNoRefresh", {} ),
--  x_newtSetColors                 = define_c_proc( libnewt, "newtSetColors", {C_POINTER} ),
    x_newtSetColor                  = define_c_proc( libnewt, "newtSetColor", {C_INT,C_POINTER,C_POINTER} ),
    x_newtRefresh                   = define_c_proc( libnewt, "newtRefresh", {} ),
    x_newtSuspend                   = define_c_proc( libnewt, "newtSuspend", {} ),
    x_newtSetSuspendCallback        = define_c_proc( libnewt, "newtSetSuspendCallback", {C_POINTER,C_POINTER} ),
    x_newtSetHelpCallback           = define_c_proc( libnewt, "newtSetHelpCallback", {C_POINTER} ),
    x_newtResume                    = define_c_func( libnewt, "newtResume", {}, C_INT ),
    x_newtPushHelpLine              = define_c_proc( libnewt, "newtPushHelpLine", {C_POINTER} ),
    x_newtRedrawHelpLine            = define_c_proc( libnewt, "newtRedrawHelpLine", {} ),
    x_newtPopHelpLine               = define_c_proc( libnewt, "newtPopHelpLine", {} ),
    x_newtDrawRootText              = define_c_proc( libnewt, "newtDrawRootText", {C_INT,C_INT,C_POINTER} ),
    x_newtBell                      = define_c_proc( libnewt, "newtBell", {} ),
    x_newtCursorOff                 = define_c_proc( libnewt, "newtCursorOff", {} ),
    x_newtCursorOn                  = define_c_proc( libnewt, "newtCursorOn", {} ),
    x_newtCompactButton             = define_c_func( libnewt, "newtCompactButton", {C_INT,C_INT,C_POINTER}, C_POINTER ),
    x_newtButton                    = define_c_func( libnewt, "newtButton", {C_INT,C_INT,C_POINTER}, C_POINTER ),
    x_newtCheckbox                  = define_c_func( libnewt, "newtCheckbox", {C_INT,C_INT,C_POINTER,C_CHAR,C_POINTER,C_POINTER}, C_POINTER ),
    x_newtCheckboxGetValue          = define_c_func( libnewt, "newtCheckboxGetValue", {C_POINTER}, C_CHAR ),
    x_newtCheckboxSetValue          = define_c_proc( libnewt, "newtCheckboxSetValue", {C_POINTER,C_CHAR} ),
    x_newtCheckboxSetFlags          = define_c_proc( libnewt, "newtCheckboxSetFlags", {C_POINTER,C_INT,C_INT} ),
    x_newtRadiobutton               = define_c_func( libnewt, "newtRadiobutton", {C_INT,C_INT,C_POINTER,C_INT,C_POINTER}, C_POINTER ),
    x_newtRadioGetCurrent           = define_c_func( libnewt, "newtRadioGetCurrent", {C_POINTER}, C_POINTER ),
    x_newtGetScreenSize             = define_c_proc( libnewt, "newtGetScreenSize", {C_POINTER,C_POINTER} ),
    x_newtLabel                     = define_c_func( libnewt, "newtLabel", {C_INT,C_INT,C_POINTER}, C_POINTER ),
    x_newtLabelSetText              = define_c_proc( libnewt, "newtLabelSetText", {C_POINTER,C_POINTER} ),
    x_newtVerticalScrollbar         = define_c_func( libnewt, "newtVerticalScrollbar", {C_INT,C_INT,C_INT,C_INT,C_INT}, C_POINTER ),
    x_newtScrollbarSet              = define_c_proc( libnewt, "newtScrollbarSet", {C_POINTER,C_INT,C_INT} ),
    x_newtListbox                   = define_c_func( libnewt, "newtListbox", {C_INT,C_INT,C_INT,C_INT}, C_POINTER ),
    x_newtListboxGetCurrent         = define_c_func( libnewt, "newtListboxGetCurrent", {C_POINTER}, C_POINTER ),
    x_newtListboxSetCurrent         = define_c_proc( libnewt, "newtListboxSetCurrent", {C_POINTER,C_INT} ),
    x_newtListboxSetCurrentByKey    = define_c_proc( libnewt, "newtListboxSetCurrentByKey", {C_POINTER,C_POINTER} ),
    x_newtListboxSetEntry           = define_c_proc( libnewt, "newtListboxSetEntry", {C_POINTER,C_INT,C_POINTER} ),
    x_newtListboxSetWidth           = define_c_proc( libnewt, "newtListboxSetWidth", {C_POINTER,C_INT} ),
    x_newtListboxSetData            = define_c_proc( libnewt, "newtListboxSetData", {C_POINTER,C_INT,C_POINTER} ),
    x_newtListboxAppendEntry        = define_c_func( libnewt, "newtListboxAppendEntry", {C_POINTER,C_POINTER,C_POINTER}, C_INT ),
    x_newtListboxInsertEntry        = define_c_func( libnewt, "newtListboxInsertEntry", {C_POINTER,C_POINTER,C_POINTER,C_POINTER}, C_INT ),
    x_newtListboxDeleteEntry        = define_c_func( libnewt, "newtListboxDeleteEntry", {C_POINTER,C_POINTER}, C_INT ),
    x_newtListboxClear              = define_c_proc( libnewt, "newtListboxClear", {C_POINTER} ),
    x_newtListboxGetEntry           = define_c_proc( libnewt, "newtListboxGetEntry", {C_POINTER,C_INT,C_POINTER,C_POINTER} ),
    x_newtListboxGetSelection       = define_c_func( libnewt, "newtListboxGetSelection", {C_POINTER,C_POINTER}, C_POINTER ),
    x_newtListboxClearSelection     = define_c_proc( libnewt, "newtListboxClearSelection", {C_POINTER} ),
    x_newtListboxSelectItem         = define_c_proc( libnewt, "newtListboxSelectItem", {C_POINTER,C_POINTER,C_INT} ),
    x_newtListboxItemCount          = define_c_func( libnewt, "newtListboxItemCount", {C_POINTER}, C_INT ),
    x_newtTextboxReflowed           = define_c_func( libnewt, "newtTextboxReflowed", {C_INT,C_INT,C_POINTER,C_INT,C_INT,C_INT,C_INT}, C_POINTER ),
    x_newtTextbox                   = define_c_func( libnewt, "newtTextbox", {C_INT,C_INT,C_INT,C_INT,C_INT}, C_POINTER ),
    x_newtTextboxSetText            = define_c_proc( libnewt, "newtTextboxSetText", {C_POINTER,C_POINTER} ),
    x_newtTextboxSetHeight          = define_c_proc( libnewt, "newtTextboxSetHeight", {C_POINTER,C_INT} ),
    x_newtTextboxGetNumLines        = define_c_func( libnewt, "newtTextboxGetNumLines", {C_POINTER}, C_INT ),
    x_newtReflowText                = define_c_func( libnewt, "newtReflowText", {C_POINTER,C_INT,C_INT,C_INT,C_POINTER,C_POINTER}, C_POINTER )

public enum NEWT_EXIT_HOTKEY = 0, NEWT_EXIT_COMPONENT, NEWT_EXIT_FDREADY,
    NEWT_EXIT_TIMER

public enum type newtExitStruct
    ExitStruct_reason = 0,
    ExitStruct_u_watch = 4,
    ExitStruct_u_key = 4,
    ExitStruct_u_co = 4
end type

public constant
    x_newtForm                      = define_c_func( libnewt, "newtForm", {C_POINTER,C_POINTER,C_INT}, C_POINTER ),
    x_newtFormSetTimer              = define_c_proc( libnewt, "newtFormSetTimer", {C_POINTER,C_INT} ),
    x_newtFormWatchFd               = define_c_proc( libnewt, "newtFormWatchFd", {C_POINTER,C_INT,C_INT} ),
    x_newtFormSetSize               = define_c_proc( libnewt, "newtFormSetSize", {C_POINTER} ),
    x_newtFormGetCurrent            = define_c_func( libnewt, "newtFormGetCurrent", {C_POINTER}, C_POINTER ),
    x_newtFormSetBackground         = define_c_proc( libnewt, "newtFormSetBackground", {C_POINTER,C_INT} ),
    x_newtFormSetCurrent            = define_c_proc( libnewt, "newtFormSetCurrent", {C_POINTER,C_POINTER} ),
    x_newtFormAddComponent          = define_c_proc( libnewt, "newtFormAddComponent", {C_POINTER,C_POINTER} ),
    x_newtFormAddComponents         = define_c_proc( libnewt, "newtFormAddComponents", {C_POINTER /* ... */ } ),
    x_newtFormSetHeight             = define_c_proc( libnewt, "newtFormSetHeight", {C_POINTER,C_INT} ),
    x_newtFormSetWidth              = define_c_proc( libnewt, "newtFormSetWidth", {C_POINTER,C_INT} ),
    x_newtRunForm                   = define_c_func( libnewt, "newtRunForm", {C_POINTER}, C_POINTER ), /* obsolete */
    x_newtFormRun                   = define_c_proc( libnewt, "newtFormRun", {C_POINTER,C_POINTER} ),
    x_newtDrawForm                  = define_c_proc( libnewt, "newtDrawForm", {C_POINTER} ),
    x_newtFormAddHotKey             = define_c_proc( libnewt, "newtFormAddHotKey", {C_POINTER,C_INT} )

public type newtEntryFilter( object x ) return atom(x) end type

public constant
    x_newtEntry                     = define_c_func( libnewt, "newtEntry", {C_INT,C_INT,C_POINTER,C_INT,C_POINTER,C_INT}, C_POINTER ),
    x_newtEntrySet                  = define_c_proc( libnewt, "newtEntrySet", {C_POINTER,C_POINTER,C_INT} ),
    x_newtEntrySetFilter            = define_c_proc( libnewt, "newtEntrySetFilter", {C_POINTER,C_POINTER,C_POINTER} ),
    x_newtEntryGetValue             = define_c_func( libnewt, "newtEntryGetValue", {C_POINTER}, C_POINTER ),
    x_newtEntrySetFlags             = define_c_proc( libnewt, "newtEntrySetFlags", {C_POINTER,C_INT,C_INT} ),
    x_newtScale                     = define_c_func( libnewt, "newtScale", {C_INT,C_INT,C_INT,C_LONG,C_LONG}, C_POINTER ),
    x_newtScaleSet                  = define_c_proc( libnewt, "newtScaleSet", {C_POINTER,C_ULONG,C_ULONG} ),
    x_newtComponentAddCallback      = define_c_proc( libnewt, "newtComponentAddCallback", {C_POINTER,C_POINTER,C_POINTER} ),
    x_newtComponentTakesFocus       = define_c_proc( libnewt, "newtComponentTakesFocus", {C_POINTER,C_INT} ),
    x_newtComponentAddDestroyCallback = define_c_proc( libnewt, "newtComponentAddDestroyCallback",{C_POINTER,C_POINTER,C_POINTER} ),
    x_newtFormDestroy               = define_c_proc( libnewt, "newtFormDestroy", {C_POINTER} ),
    x_newtComponentDestroy          = define_c_proc( libnewt, "newtComponentDestroy", {C_POINTER} )

public constant
    NEWT_KEY_TAB		= '\t',
    NEWT_KEY_ENTER		= '\r',
--  NEWT_KEY_SUSPEND		= #032, -- '\032'	/* Ctrl-Z */
    NEWT_KEY_ESCAPE		= 27,
    NEWT_KEY_EXTRA_BASE		= #8000,
    NEWT_KEY_UP			= NEWT_KEY_EXTRA_BASE + 1,
    NEWT_KEY_DOWN		= NEWT_KEY_EXTRA_BASE + 2,
    NEWT_KEY_LEFT		= NEWT_KEY_EXTRA_BASE + 4,
    NEWT_KEY_RIGHT		= NEWT_KEY_EXTRA_BASE + 5,
    NEWT_KEY_BKSPC		= NEWT_KEY_EXTRA_BASE + 6,
    NEWT_KEY_DELETE		= NEWT_KEY_EXTRA_BASE + 7,
    NEWT_KEY_HOME		= NEWT_KEY_EXTRA_BASE + 8,
    NEWT_KEY_END		= NEWT_KEY_EXTRA_BASE + 9,
    NEWT_KEY_UNTAB		= NEWT_KEY_EXTRA_BASE + 10,
    NEWT_KEY_PGUP		= NEWT_KEY_EXTRA_BASE + 11,
    NEWT_KEY_PGDN		= NEWT_KEY_EXTRA_BASE + 12,
    NEWT_KEY_INSERT		= NEWT_KEY_EXTRA_BASE + 13,
    NEWT_KEY_F1			= NEWT_KEY_EXTRA_BASE + 101,
    NEWT_KEY_F2			= NEWT_KEY_EXTRA_BASE + 102,
    NEWT_KEY_F3			= NEWT_KEY_EXTRA_BASE + 103,
    NEWT_KEY_F4			= NEWT_KEY_EXTRA_BASE + 104,
    NEWT_KEY_F5			= NEWT_KEY_EXTRA_BASE + 105,
    NEWT_KEY_F6			= NEWT_KEY_EXTRA_BASE + 106,
    NEWT_KEY_F7			= NEWT_KEY_EXTRA_BASE + 107,
    NEWT_KEY_F8			= NEWT_KEY_EXTRA_BASE + 108,
    NEWT_KEY_F9			= NEWT_KEY_EXTRA_BASE + 109,
    NEWT_KEY_F10		= NEWT_KEY_EXTRA_BASE + 110,
    NEWT_KEY_F11		= NEWT_KEY_EXTRA_BASE + 111,
    NEWT_KEY_F12		= NEWT_KEY_EXTRA_BASE + 112,
    NEWT_KEY_RESIZE		= NEWT_KEY_EXTRA_BASE + 113,
    NEWT_ANCHOR_LEFT		= #0001, -- (1 << 0)
    NEWT_ANCHOR_RIGHT		= #0002, -- (1 << 1)
    NEWT_ANCHOR_TOP		= #0004, -- (1 << 2)
    NEWT_ANCHOR_BOTTOM		= #0008, -- (1 << 3)
    NEWT_GRID_FLAG_GROWX	= #0001, -- (1 << 0)
    NEWT_GRID_FLAG_GROWY	= #0002  -- (1 << 1)

public type newtGrid( object x ) return atom(x) end type

public enum type newtGridElement NEWT_GRID_EMPTY = 0,
    NEWT_GRID_COMPONENT, NEWT_GRID_SUBGRID
end type

public constant
    x_newtCreateGrid                = define_c_func( libnewt, "newtCreateGrid", {C_INT,C_INT}, C_POINTER ),
    x_newtGridVStacked              = define_c_func( libnewt, "newtGridVStacked", {C_INT,C_POINTER /* ... */}, C_POINTER ),
    x_newtGridVCloseStacked         = define_c_func( libnewt, "newtGridVCloseStacked", {C_INT,C_POINTER /* ... */}, C_POINTER ),
    x_newtGridHStacked              = define_c_func( libnewt, "newtGridHStacked", {C_INT,C_POINTER /* ... */}, C_POINTER ),
    x_newtGridHCloseStacked         = define_c_func( libnewt, "newtGridHCloseStacked", {C_INT,C_POINTER /* ... */}, C_POINTER ),
    x_newtGridBasicWindow           = define_c_func( libnewt, "newtGridBasicWindow", {C_POINTER,C_POINTER,C_POINTER}, C_POINTER ),
    x_newtGridSimpleWindow          = define_c_func( libnewt, "newtGridSimpleWindow", {C_POINTER,C_POINTER,C_POINTER}, C_POINTER ),
    x_newtGridSetField              = define_c_proc( libnewt, "newtGridSetField", {C_POINTER,C_INT,C_INT,C_INT,C_POINTER,C_INT,C_INT,C_INT,C_INT,C_INT,C_INT} ),
    x_newtGridPlace                 = define_c_proc( libnewt, "newtGridPlace", {C_POINTER,C_INT,C_INT} ),
    x_newtGridFree                  = define_c_proc( libnewt, "newtGridFree", {C_POINTER,C_INT} ),
    x_newtGridGetSize               = define_c_proc( libnewt, "newtGridGetSize", {C_POINTER,C_POINTER,C_POINTER} ),
    x_newtGridWrappedWindow         = define_c_proc( libnewt, "newtGridWrappedWindow", {C_POINTER,C_POINTER} ),
    x_newtGridWrappedWindowAt       = define_c_proc( libnewt, "newtGridWrappedWindowAt", {C_POINTER,C_POINTER,C_INT,C_INT} ),
    x_newtGridAddComponentsToForm   = define_c_proc( libnewt, "newtGridAddComponentsToForm", {C_POINTER,C_POINTER,C_INT} ),
    x_newtButtonBar                 = define_c_func( libnewt, "newtButtonBarv", {C_POINTER,C_POINTER /* ... */}, C_POINTER ),
    x_newtWinMessage                = define_c_proc( libnewt, "newtWinMessage", {C_POINTER,C_POINTER,C_POINTER /* ... */} ),
    x_newtWinChoice                 = define_c_func( libnewt, "newtWinChoice", {C_POINTER,C_POINTER,C_POINTER,C_POINTER /* ... */}, C_INT ),
    x_newtWinTernary                = define_c_func( libnewt, "newtWinTernary", {C_POINTER,C_POINTER,C_POINTER,C_POINTER,C_POINTER /* ... */}, C_INT )

public enum type newtWinEntry
    WinEntry_text = 0,
    WinEntry_value = 4,
    WinEntry_flags = 8
end type

public constant
    x_newtWinEntries                = define_c_func( libnewt, "newtWinEntries", {C_POINTER,C_POINTER,C_INT,C_INT,C_INT,C_INT,C_POINTER,C_POINTER /* ... */}, C_INT )

public function Init( integer clear = 0 )
    integer retval = c_func( x_newtInit, {} )
    if clear then c_proc( x_newtCls, {} ) end if
    return retval
end function

public function Finished( integer clear = 0 )
    integer retval = c_func( x_newtFinished, {} )
    if clear then clear_screen() end if
    return retval
end function

public procedure Cls()
    c_proc( x_newtCls, {} )
end procedure

public procedure ResizeScreen( integer redraw )
    c_proc( x_newtResizeScreen, {redraw} )
end procedure

public procedure WaitForKey()
    c_proc( x_newtWaitForKey, {} )
end procedure

public procedure ClearKeyBuffer()
    c_proc( x_newtClearKeyBuffer, {} )
end procedure

public procedure Delay( integer usecs )
    c_proc( x_newtDelay, {usecs} )
end procedure

public function OpenWindow( integer left, integer top, integer width, integer height, object title = 0 )
    if sequence( title ) then title = allocate_string( title, 1 ) end if
    return c_func( x_newtOpenWindow, {left,top,width,height,title} )
end function

public function CenteredWindow( integer width, integer height, object title = 0 )
    if sequence( title ) then title = allocate_string( title, 1 ) end if
    return c_func( x_newtCenteredWindow, {width,height,title} )
end function

public procedure PopWindow( integer refresh = 1 )
    if refresh then c_proc( x_newtPopWindow, {} )
    else c_proc( x_newtPopWindowNoRefresh, {} ) end if
end procedure

public procedure PopWindowNoRefresh()
    c_proc( x_newtPopWindowNoRefresh, {} )
end procedure

--public procedure SetColors( sequence colors )
--  c_proc( x_newtSetColors, {/* ??? */} )
--end procedure

public procedure SetColor( integer colorset, integer fg, integer bg )
    atom pfg = allocate( 1, 1 ) poke( pfg, and_bits(fg, #FF) )
    atom pbg = allocate( 1, 1 ) poke( pbg, and_bits(bg, #FF) )
    c_proc( x_newtSetColor, {colorset,pfg,pbg} )
end procedure

public procedure Refresh()
    c_proc( x_newtRefresh, {} )
end procedure

public procedure Suspend()
    c_proc( x_newtSuspend, {} )
end procedure

public procedure SetSuspendCallback( atom cb, atom data = 0 )
    c_proc( x_newtSetSuspendCallback, {cb,data} )
end procedure

public procedure SetSuspendRoutine( integer rid, atom data = 0 )
    atom cb = call_back( rid )
    c_proc( x_newtSetSuspendCallback, {cb,data} )
end procedure

public procedure SetHelpCallback( atom cb )
    c_proc( x_newtSetHelpCallback, {cb} )
end procedure

public procedure SetHelpRoutine( integer rid )
    atom cb = call_back( rid )
    c_proc( x_newtSetHelpCallback, {cb} )
end procedure

public function Resume()
    return c_func( x_newtResume, {} )
end function

public procedure PushHelpLine( object text = 0 )
    if sequence( text ) then text = allocate_string( text, 1 ) end if
    c_proc( x_newtPushHelpLine, {text} )
end procedure

public procedure RedrawHelpLine()
    c_proc( x_newtRedrawHelpLine, {} )
end procedure

public procedure PopHelpLine()
    c_proc( x_newtPopHelpLine, {} )
end procedure

public procedure DrawRootText( integer col, integer row, object text )
    if sequence( text ) then text = allocate_string( text, 1 ) end if
    c_proc( x_newtDrawRootText, {col,row,text} )
end procedure

public procedure Bell()
    c_proc( x_newtBell, {} )
end procedure

public procedure CursorOff()
    c_proc( x_newtCursorOff, {} )
end procedure

public procedure CursorOn()
    c_proc( x_newtCursorOn, {} )
end procedure

public function CompactButton( integer left, integer top, object text )
    if sequence( text ) then text = allocate_string( text, 1 ) end if
    return c_func( x_newtCompactButton, {left,top,text} )
end function

public function Button( integer left, integer top, object text )
    if sequence( text ) then text = allocate_string( text, 1 ) end if
    return c_func( x_newtButton, {left,top,text} )
end function

public function Checkbox( integer left, integer top, object text, integer defValue = ' ', object seq = 0, atom result = 0 )
    if sequence( text ) then text = allocate_string( text, 1 ) end if
    if sequence( seq ) then seq = allocate_string( seq, 1 ) end if
    return c_func( x_newtCheckbox, {left,top,text,defValue,seq,result} )
end function

public function CheckboxGetValue( newtComponent co )
    return c_func( x_newtCheckboxGetValue, {co} )
end function

public procedure CheckboxSetValue( newtComponent co, integer value )
    c_proc( x_newtCheckboxSetValue, {co,value} )
end procedure

public procedure CheckboxSetFlags( newtComponent co, integer flags = 0, newtFlagsSense sense = NEWT_FLAGS_SET )
    c_proc( x_newtCheckboxSetFlags, {co,flags,sense} )
end procedure

public function Radiobutton( integer left, integer top, object text, integer isDefault = 0, newtComponent prevButton = 0 )
    if sequence( text ) then text = allocate_string( text, 1 ) end if
    return c_func( x_newtRadiobutton, {left,top,text,isDefault,prevButton} )
end function

public function RadioGetCurrent( newtComponent setMember )
    return c_func( x_newtRadioGetCurrent, {setMember} )
end function

public function GetScreenSize()
    atom ptr = allocate( 8 )
    c_proc( x_newtGetScreenSize, {ptr+0,ptr+4} )
    integer cols = peek4s( ptr+0 )
    integer rows = peek4s( ptr+4 )
    free( ptr )
    return {cols, rows}
end function

public function Label( integer left, integer top, object text )
    if sequence( text ) then text = allocate_string( text, 1 ) end if
    return c_func( x_newtLabel, {left,top,text} )
end function

public function LabelSetText( newtComponent co, object text )
    if sequence( text ) then text = allocate_string( text, 1 ) end if
    return c_func( x_newtLabelSetText, {co,text} )
end function

public function VerticalScrollbar( integer left, integer top, integer height, integer normalColorset = 0, integer thumbColorset = 0 )
    return c_func( x_newtVerticalScrollbar, {left,top,height,normalColorset,thumbColorset} )
end function

public function ScrollbarSet( newtComponent co, integer where, integer total )
    return c_func( x_newtScrollbarSet, {co,where,total} )
end function

public function Listbox( integer left, integer top, integer height, integer flags = 0 )
    return c_func( x_newtListbox, {left,top,height,flags} )
end function

public function ListboxGetCurrent( newtComponent co )
    return c_func( x_newtListboxGetCurrent, {co} )
end function

public procedure ListboxSetCurrent( newtComponent co, integer num )
    c_proc( x_newtListboxSetCurrent, {co,num} )
end procedure

public procedure ListboxSetCurrentByKey( newtComponent co, atom key )
    c_proc( x_newtListboxSetCurrentByKey, {co,key} )
end procedure

public procedure ListboxSetEntry( newtComponent co, integer num, object text )
    if sequence( text ) then text = allocate_string( text, 1 ) end if
    c_proc( x_newtListboxSetEntry, {co,num,text} )
end procedure

public procedure ListboxSetWidth( newtComponent co, integer width )
    c_proc( x_newtListboxSetWidth, {co,width} )
end procedure

public procedure ListboxSetData( newtComponent co, integer num, atom data )
    c_proc( x_newtListboxSetData, {co,num,data} )
end procedure

public function ListboxAppendEntry( newtComponent co, object text, atom data )
    if sequence( text ) then text = allocate_string( text, 1 ) end if
    return c_func( x_newtListboxAppendEntry, {co,text,data} )
end function

public function ListboxAppendEntries( newtComponent co, sequence entries, object data = 0 )
    integer count = length( entries )
    sequence result = repeat( 0, count )
    if atom( data ) then data = repeat( data, count ) end if
    for i = 1 to count do
        if sequence( entries[i] ) then entries[i] = allocate_string( entries[i], 1 ) end if
        result[i] = c_func( x_newtListboxAppendEntry, {co,entries[i],data[i]} )
    end for
    return result
end function

public function ListboxInsertEntry( newtComponent co, object text, atom data, atom key )
    if sequence( text ) then text = allocate_string( text, 1 ) end if
    return c_func( x_newtListboxInsertEntry, {co,text,data,key} )
end function

public function ListboxDeleteEntry( newtComponent co, atom data )
    return c_func( x_newtListboxDeleteEntry, {co,data} )
end function

public procedure ListboxClear( newtComponent co )
    c_proc( x_newtListboxClear, {co} )
end procedure

public function ListboxGetEntry( newtComponent co, integer num )
    atom ptext = allocate( 4 )
    atom pdata = allocate( 4 )
    c_proc( x_newtListboxGetEntry, {co,num,ptext,pdata} )
    sequence text = peek_string( ptext )
    atom data = peek4u( pdata )
    free( ptext )
    free( pdata )
    return {text, data}
end function

public function ListboxGetSelection( newtComponent co )
    atom pnumitems = allocate( 4 )
    atom pdata = c_func( x_newtListboxGetSelection, {co,pnumitems} )
    integer numitems = peek4s( pnumitems )
    sequence items = repeat( 0, numitems )
    for i = 1 to numitems do
        items[i] = peek4u( pdata + (numitems-1) * 4 )
    end for
    free( pnumitems )
    return items
end function

public procedure ListboxClearSelection( newtComponent co )
    c_proc( x_newtListboxClearSelection, {co} )
end procedure

public procedure ListboxSelectItem( newtComponent co, atom key, newtFlagsSense sense = NEWT_FLAGS_SET )
    c_proc( x_newtListboxSelectItem, {co,key,sense} )
end procedure

public function ListboxItemCount( newtComponent co )
    return c_func( x_newtListboxItemCount, {co} )
end function

public function TextboxReflowed( integer left, integer top, object text, integer width, integer flexDown, integer flexUp, integer flags = 0 )
    if sequence( text ) then text = allocate_string( text, 1 ) end if
    return c_func( x_newtTextboxReflowed, {left,top,text,width,flexDown,flexUp,flags} )
end function

public function Textbox( integer left, integer top, integer width, integer height, integer flags = 0 )
    return c_func( x_newtTextbox, {left,top,width,height,flags} )
end function

public procedure TextboxSetText( newtComponent co, object text )
    if sequence( text ) then text = allocate_string( text, 1 ) end if
    c_proc( x_newtTextboxSetText, {co,text} )
end procedure

public procedure TextboxSetHeight( newtComponent co, integer height )
    c_proc( x_newtTextboxSetHeight, {co,height} )
end procedure

public function TextboxGetNumLines( newtComponent co )
    return c_func( x_newtTextboxGetNumLines, {co} )
end function

public function ReflowText( object text, integer width, integer flexDown, integer flexUp )
    if sequence( text ) then text = allocate_string( text, 1 ) end if
    atom ptr = allocate( 8 )
    atom presult = c_func( x_newtReflowText, {text,width,flexDown,flexUp,ptr+0,ptr+4} )
    integer actualWidth = peek4s( ptr+0 )
    integer actualHeight = peek4s( ptr+4 )
    free( ptr )
    return peek_string( presult )
end function

public function Form( newtComponent vertBar = 0, atom helpTag = 0, integer flags = 0 )
    return c_func( x_newtForm, {vertBar,helpTag,flags} )
end function

public procedure FormSetTimer( newtComponent form, integer millisecs )
    c_proc( x_newtFormSetTimer, {form,millisecs} )
end procedure

public procedure FormWatchFd( newtComponent form, integer fd, integer fdFlags = NEWT_FD_READ )
    c_proc( x_newtFormWatchFd, {form,fd,fdFlags} )
end procedure

public procedure FormSetSize( newtComponent co )
    c_proc( x_newtFormSetSize, {co} )
end procedure

public function FormGetCurrent( newtComponent co )
    return c_func( x_newtFormGetCurrent, {co} )
end function

public procedure FormSetBackground( newtComponent co, integer color )
    c_proc( x_newtFormSetBackground, {co,color} )
end procedure

public procedure FormSetCurrent( newtComponent co, newtComponent subco )
    c_proc( x_newtFormSetCurrent, {co,subco} )
end procedure

public procedure FormAddComponent( newtComponent form, newtComponent co )
    c_proc( x_newtFormAddComponent, {form,co} )
end procedure

public procedure FormAddComponents( newtComponent form, sequence co )
    for i = 1 to length( co ) do
        c_proc( x_newtFormAddComponent, {form,co[i]} )
    end for
end procedure

public procedure FormSetHeight( newtComponent co, integer height )
    c_proc( x_newtFormSetHeight, {co,height} )
end procedure

public procedure FormSetWidth( newtComponent co, integer width )
    c_proc( x_newtFormSetHeight, {co,width} )
end procedure

public function RunForm( newtComponent form )
    return c_func( x_newtRunForm, {form} )
end function

public function FormRun( newtComponent co )
    atom es = allocate( 8 )
    c_proc( x_newtFormRun, {co,es} )
    integer reason = peek4u( es + ExitStruct_reason )
    sequence result = {reason, -1}
    switch reason do
        case NEWT_EXIT_HOTKEY then
            result[2] = peek4u( es + ExitStruct_u_key )
        case NEWT_EXIT_COMPONENT then
            result[2] = peek4s( es + ExitStruct_u_co )
        case NEWT_EXIT_FDREADY then
            result[2] = peek4u( es + ExitStruct_u_watch )
    end switch
    free( es )
    return result
end function

public procedure DrawForm( newtComponent co )
    c_proc( x_newtDrawForm, {co} )
end procedure

public procedure FormAddHotKey( newtComponent co, integer key )
    c_proc( x_newtFormAddHotKey, {co,key} )
end procedure

public function Entry( integer left, integer top, object initialValue, integer width, atom resultPtr = 0, integer flags = 0 )
    if sequence( initialValue ) then initialValue = allocate_string( initialValue, 1 ) end if
    return c_func( x_newtEntry, {left,top,initialValue,width,resultPtr,flags} )
end function

public procedure EntrySet( newtComponent co, object value, integer cursorAtEnd = 0 )
    c_proc( x_newtEntrySet, {co,value,cursorAtEnd} )
end procedure

public procedure EntrySetFilter( newtComponent co, newtEntryFilter filter, atom data )
    c_proc( x_newtEntrySetFilter, {co,filter,data} )
end procedure

public function EntryGetValue( newtComponent co )
    atom value = c_func( x_newtEntryGetValue, {co} )
    return peek_string( value )
end function

public procedure EntrySetFlags( newtComponent co, integer flags, newtFlagsSense sense = NEWT_FLAGS_SET )
    c_proc( x_newtEntrySetFlags, {co,flags,sense} )
end procedure

public function Scale( integer left, integer top, integer width, atom fullValue )
    -- stupid 64-bit integers :-/
    atom dwLow = and_bits( fullValue, #FFFFFFFF )
    atom dwHigh = and_bits( fullValue / #100000000, #FFFFFFFF )
    return c_func( x_newtScale, {left,top,width,dwLow,dwHigh} )
end function

public procedure ScaleSet( newtComponent co, atom amount )
    atom dwLow = and_bits( amount, #FFFFFFFF )
    atom dwHigh = and_bits( amount / #100000000, #FFFFFFFF )
    c_proc( x_newtScaleSet, {co,amount} )
end procedure

public procedure ComponentAddCallback( newtComponent co, newtCallback f, atom data )
    c_proc( x_newtComponentAddCallback, {co,f,data} )
end procedure

public procedure ComponentTakesFocus( newtComponent co, integer val )
    c_proc( x_newtComponentTakesFocus, {co,val} )
end procedure

public procedure ComponentAddDestroyCallback( newtComponent co, newtCallback f, atom data )
    c_proc( x_newtComponentAddDestroyCallback, {co,f,data} )
end procedure

public procedure FormDestroy( newtComponent form )
    c_proc( x_newtFormDestroy, {form} )
end procedure

public procedure ComponentDestroy( newtComponent co )
    c_proc( x_newtComponentDestroy, {co} )
end procedure

-- from newt-0.52.10/grid.c  (well, sort of..)
function stackem( integer isVert, sequence items, integer close )

    integer cols = 1, rows = 1, count = length( items )

    if isVert then rows = count else cols = count end if

    newtComponent grid = c_func( x_newtCreateGrid, {cols, rows} )

    integer col = 0, row = 0, left = 0, top = 0

    for i = 1 to count do

        if isVert then row = i - 1 else col = i - 1 end if

        sequence item = items[i]

        if i > 1 and not close then
            if isVert then left = 1 else top = 1 end if
        end if

        c_proc( x_newtGridSetField, {grid,col,row,item[1],item[2],left,top,0,0,0,0} )

    end for

    return grid
end function

public function CreateGrid( integer cols, integer rows )
    return c_func( x_newtCreateGrid, {cols,rows} )
end function

public function GridVStacked( sequence items )
    return stackem( 1, items, 0 )
end function

public function GridVCloseStacked( sequence items )
    return stackem( 1, items, 1 )
end function

public function GridHStacked( sequence items )
    return stackem( 0, items, 0 )
end function

public function GridHCloseStacked( sequence items )
    return stackem( 0, items, 1 )
end function

public function GridBasicWindow( newtComponent text, newtGrid middle, newtGrid buttons )
    return c_func( x_newtGridBasicWindow, {text,middle,buttons} )
end function

public function GridSimpleWindow( newtComponent text, newtGrid middle, newtGrid buttons )
    return c_func( x_newtGridSimpleWindow, {text,middle,buttons} )
end function

public procedure GridSetField( newtGrid grid, integer col, integer row, newtGridElement itemType, atom itemValue, integer padLeft = 0,
        integer padTop = 0, integer padRight = 0, integer padBottom = 0, integer anchor = 0, integer flags = 0 )
    c_proc( x_newtGridSetField, {grid,col,row,itemType,itemValue,padLeft,padTop,padRight,padBottom,anchor,flags} )
end procedure

public procedure GridPlace( newtGrid grid, integer left, integer top )
    c_proc( x_newtGridPlace, {grid,left,top} )
end procedure

public procedure GridDestroy( newtGrid grid, integer recurse = 1 )
    c_proc( x_newtGridFree, {grid,recurse} )
end procedure

public procedure GridFree( newtGrid grid, integer recurse = 1 )
    c_proc( x_newtGridFree, {grid,recurse} )
end procedure

public function GridGetSize( newtGrid grid )
    atom ptr = allocate( 8 )
    c_proc( x_newtGridGetSize, {ptr+0,ptr+4} )
    integer width = peek4s( ptr+0 )
    integer height = peek4s( ptr+4 )
    free( ptr )
    return {width, height}
end function

public procedure GridWrappedWindow( newtGrid grid, object title = 0, integer left = 0, integer top = 0 )
    if sequence( title ) then title = allocate_string( title, 1 ) end if
    if left = 0 and top = 0 then
        c_proc( x_newtGridWrappedWindow, {grid,title} )
    else
        c_proc( x_newtGridWrappedWindowAt, {grid,title,left,top} )
    end if
end procedure

public procedure GridAddComponentsToForm( newtGrid grid, newtComponent form, integer recurse = 1 )
    c_proc( x_newtGridAddComponentsToForm, {grid,form,recurse} )
end procedure

public function ButtonBar( sequence items )

    integer count = length( items )
    sequence buttons = repeat( 0, count )

    newtGrid grid = c_func( x_newtCreateGrid, {count,1} )

    for i = 1 to count do

        if sequence( items[i] ) then items[i] = allocate_string( items[i], 1 ) end if
        buttons[i] = c_func( x_newtButton, {-1,-1,items[i]} )

        integer left = 1 if i = 1 then left = 0 end if
        c_proc( x_newtGridSetField, {grid,i-1,0,NEWT_GRID_COMPONENT,buttons[i],left,0,0,0,0,0} )

    end for

    return {grid, buttons}
end function

public procedure WinMessage( object title, object buttonText, object text )
    if sequence( title ) then title = allocate_string( title, 1 ) end if
    if sequence( buttonText ) then buttonText = allocate_string( buttonText, 1 ) end if
    if sequence( text ) then text = allocate_string( text, 1 ) end if
    c_proc( x_newtWinMessage, {title,buttonText,text} )
end procedure

public function WinChoice( object title, object text, object button1, object button2, object button3 = 0 )
    if sequence( title ) then title = allocate_string( title, 1 ) end if
    if sequence( text ) then text = allocate_string( text, 1 ) end if
    if sequence( button1 ) then button1 = allocate_string( button1, 1 ) end if
    if sequence( button2 ) then button2 = allocate_string( button2, 1 ) end if
    if sequence( button3 ) then button3 = allocate_string( button3, 1 ) end if
    if equal( button3, 0 ) then
        return c_func( x_newtWinChoice, {title,button1,button2,text} )
    else
        return c_func( x_newtWinTernary, {title,button1,button2,button3,text} )
    end if
end function

